import os
import signal
import sys

from omniduct.utils.config import config as omniduct_config
from omniduct.utils.debug import logger

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
    from subprocess32 import TimeoutExpired
else:
    import subprocess
    from subprocess import TimeoutExpired

__all__ = ['run_in_subprocess', 'TimeoutExpired', 'Timeout', 'TimeoutError']

DEFAULT_SUBPROCESS_CONFIG = {
    'shell': True,
    'close_fds': False,
    'stdin': None,
    'stdout': subprocess.PIPE,
    'stderr': subprocess.PIPE,
    'preexec_fn': os.setsid  # Set the process as the group leader, so we can kill recursively
}


class SubprocessResults(object):

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


def run_in_subprocess(cmd, check_output=False, ** kwargs):
    """
    Execute command using default subprocess configuration.

    Parameters
    ----------
    cmd : string
        Command to be executed in subprocess.
    kwargs : keywords
        Options to pass to subprocess.Popen.

    Returns
    -------
    proc : Popen subprocess
        Subprocess used to run command.
    """

    logger.debug('Executing command: {0}'.format(cmd))
    config = DEFAULT_SUBPROCESS_CONFIG.copy()
    config.update(kwargs)
    if not check_output:
        if omniduct_config.logging_level < 20:
            config['stdout'] = None
            config['stderr'] = None
        else:
            config['stdout'] = open(os.devnull, 'w')
            config['stderr'] = open(os.devnull, 'w')
    timeout = config.pop('timeout', None)

    process = subprocess.Popen(cmd, **config)
    try:
        stdout, stderr = process.communicate(None, timeout=timeout)
    except subprocess.TimeoutExpired:
        os.killpg(os.getpgid(process.pid), signal.SIGINT)  # send signal to the process group, recurively killing all children
        output, unused_err = process.communicate()
        raise subprocess.TimeoutExpired(process.args, timeout, output=output)
    return SubprocessResults(returncode=process.returncode, stdout=stdout or '', stderr=stderr or '')


class TimeoutError(Exception):
    pass


class Timeout(object):

    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)



from __future__ import absolute_import, print_function

import hashlib
import logging
import os
import sys
from abc import abstractmethod

import pandas as pd
import pandas.io.sql
import six
import sqlparse
from decorator import decorator
from jinja2 import Template

from omniduct.caches.base import cached_method
from omniduct.duct import Duct
from omniduct.utils.config import config
from omniduct.utils.debug import logger, logging_scope
from omniduct.utils.magics import MagicsProvider, process_line_arguments

logging.getLogger('requests').setLevel(logging.WARNING)


config.register('date_fields',
                description='Default date fields to attempt to parse when databases',
                type=list)


@decorator
def sanitize_sqlalchemy_statement(f, self, statement, *args, **kwargs):
    try:
        from sqlalchemy.sql.base import Executable
        if isinstance(statement, Executable):
            statement = str(statement.compile(compile_kwargs={"literal_binds": True}))
    except ImportError:
        pass
    return f(self, statement, *args, **kwargs)


class DatabaseClient(Duct, MagicsProvider):
    '''
    QueryClient is an abstract class that can be subclassed to allow the databases
    of various data sources, such as databases or website apis.
    '''

    DUCT_TYPE = Duct.Type.DATABASE
    DEFAULT_PORT = None

    def __init__(self, *args, **kwargs):
        '''
        This is a shim __init__ function that passes all arguments onto
        `self._init`, which is implemented by subclasses. This allows subclasses
        to instantiate themselves with arbitrary parameters.
        '''
        Duct.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)
        self._init(*args, **kwargs)

    @abstractmethod
    def _init(self):
        pass

    def __call__(self, query, **kwargs):
        """Calls run() largely for backwards compatibility"""
        return self.query(query, **kwargs)

    # Querying
    @classmethod
    def statements_split(cls, statements):
        for statement in sqlparse.split(statements):
            statement = statement.strip()
            if statement.endswith(';'):
                statement = statement[:-1].strip()
            if statement:  # remove empty statements
                yield statement

    @classmethod
    def statement_cleanup(cls, statement):
        statement = sqlparse.format(statement, strip_comments=True, reindent=True)
        statement = os.linesep.join([line for line in statement.splitlines() if line])
        return statement

    @classmethod
    def statement_hash(cls, statement):
        statement = cls.statement_cleanup(statement)
        if sys.version_info.major == 3 or sys.version_info.major == 2 and isinstance(statement, unicode):
            statement = statement.encode('utf8')
        return hashlib.sha256(statement).hexdigest()

    @logging_scope("Query", timed=True)
    @sanitize_sqlalchemy_statement
    @cached_method(
        id_str=lambda self, kwargs: self.statement_hash(kwargs['statement']),
        use_cache=lambda self, kwargs: kwargs.pop('use_cache', True) and kwargs.get('query', False) and kwargs.get('parse', True)
    )
    def execute(self, statement, query=False, parse=True, index_field=None, date_fields=None,
                cleanup_statement=True, render_only=False, **kwargs):
        '''
        Execute a statement against the data source.

        Parameters
        ----------
        statement : The statement to be executed by the query client.
        query : Whether this statement should return data, in which case `query` should be `True`;
            and `False` otherwise.
        parse : Whether the results of this query should be converted to a pandas DataFrame.
        index_field : The field to use as an index in the dataframe, or None.
        date_fields: List of fields to be converted to datetime objects, or None.
        kwargs : Extra keyword arguments to be passed on to _execute, as implemented by subclasses.

        Returns
        -------
        A pandas.DataFrame object if `query` and `parse` are both `True`.
        A DBAPI2 cursor object if `query` is `True`, and `parse` is `False`.
        `None` otherwise.
        '''
        self.connect()
        statements = self.statements_split(statement)
        statements = [self.statement_cleanup(stmt) if cleanup_statement else stmt for stmt in statements]
        assert len(statements) > 0, "No non-empty statements were provided."
        if render_only:
            return ';\n'.join(statements)
        cursor = None
        for statement in statements[:-1]:
            cursor = self.connect()._execute(statement, query=False, cursor=cursor, **kwargs)
        cursor = self.connect()._execute(statements[-1], query, cursor=cursor, **kwargs)

        if not query or self._cursor_empty(cursor):
            return None
        if parse:
            df = self._cursor_to_dataframe(cursor)
            cursor.close()

            if date_fields is None:  # if user supplied, use as is
                date_fields = config.date_fields or []
                date_fields = [field for field in date_fields if field in df]

            if date_fields:
                try:
                    df = pandas.io.sql._parse_date_columns(df, date_fields)
                except:
                    logger.error('Unable to parse date columns. Perhaps your version of pandas is outdated.')
            if index_field is not None:
                df.set_index(index_field, inplace=True)
            return df
        else:
            return cursor

    def query(self, query, **kwargs):
        '''
        This method  is shorthand for:
        >>> client.execute(query, query=True, **kwargs)
        '''
        return self.execute(query, query=True, **kwargs)

    def query_dump(self, query, file, format='hive', **kwargs):
        '''
        This method executes a query, and streams the results into a file. This is
        especially useful for large queries, which may not fit entirely in memory.
        Currently all fields are treated as strings.

        Parameters
        ----------
        query : The query to be executed on the data source.
        file : A file-like object with a `write` method that accepts strings; or
            a string representing a path to be written to.
        format : A string indicating the format of the outputted data.
            Valid options are 'hive' and 'csv'.
        kwargs : Additional keyword arguments passed on to `self.execute`.

        Returns
        -------
        An integer representing the number of output records.
        '''
        cursor = self.execute(query, parse=False, **kwargs)
        record = cursor.fetchone()
        if format == 'hive':
            sep = "\t"
            pre = ''
            post = '\n'
        else:
            sep = '","'
            pre = '"'
            post = '"\n'
        if isinstance(file, six.string_types):
            file = open(file, 'w')
        while record is not None:
            file.write(pre + sep.join([str(v).replace('"', '\"') for v in record]) + post)
            record = cursor.fetchone()
        file.close()
        cursor.close()

    def execute_from_file(self, file, **kwargs):
        """
        Read file contents and execute them against the connected data source.

        Parameters
        ----------
        file : str
            File containing a query
        kwargs : dict
            Extra parameters to pass on to `execute`.
        """
        with open(file, 'r') as f:
            query = f.read()
        return self.query(query, **kwargs)

    def query_from_file(self, file, **kwargs):
        '''
        This method is shorthand for:
        QueryClient.execute_from_file(file, parse=True, **kwargs)
        '''
        return self.execute_file(file, parse=True, **kwargs)

    # Uploading data to data store

    def push(self, df, table, overwrite=False, **kwargs):
        '''
        This method pushes a local dataframe `df` to the connected data store
        as a table `table`.

        Parameters
        ----------
        df : pandas.DataFrame
            The pandas dataframe to push into the data store.
        table : str
            The name of the table into which the dataframe should be pushed.
        overwrite : bool
            Whether the table should be replaced, if it already exists in the data store.
        kwargs : dict
            Additional arguments which are passed on to `QueryClient._push`.
        '''
        self._push(df, table, overwrite=overwrite, **kwargs)

    @abstractmethod
    def _push(self, df, table, overwrite=True, **kwargs):
        pass

    # Table properties

    @abstractmethod
    def _execute(self, statement, query=True, cursor=None, **kwargs):
        pass

    @abstractmethod
    def _cursor_empty(self, cursor):
        pass

    def _cursor_to_dataframe(self, cursor):
        records = cursor.fetchall()
        description = cursor.description
        return pd.DataFrame(data=records, columns=[c[0] for c in description])

    def table_list(self, **kwargs):
        '''
        Return a list of table names in the data source as a DataFrame. Additional kwargs are
        passed to `QueryClient._table_list`.
        '''
        return self._table_list(**kwargs)

    @abstractmethod
    def _table_list(self, **kwargs):
        pass

    def table_exists(self, table, **kwargs):
        '''
        Return boolean if table exists in schema
        '''
        return self._table_exists(table=table, **kwargs)

    @abstractmethod
    def _table_exists(self, table, **kwargs):
        pass

    def table_desc(self, table, **kwargs):
        '''
        Describe a table in the data source. Additional kwargs are
        passed to `QueryClient._table_desc`.

        Returns a pandas dataframe of table fields and descriptors.
        '''
        return self._table_desc(table, **kwargs)

    @abstractmethod
    def _table_desc(self, table, **kwargs):
        pass

    def table_head(self, table, n=10, **kwargs):
        '''
        Show a sample of the data in `table` of the data source. `n` is the number of records
        to show in this sample. The additional `kwargs` are passed on to `QueryClient._table_head`.

        Returns a pandas DataFrame.
        '''
        return self._table_head(table, n=n, **kwargs)

    @abstractmethod
    def _table_head(self, table, n=10, **kwargs):
        pass

    def table_props(self, table, **kwargs):
        '''
        Return a dataframe of table properties for `table`.
        '''
        return self._table_props(table, **kwargs)

    @abstractmethod
    def _table_props(self, table, **kwargs):
        pass

    def _register_magics(self, base_name):
        from IPython.core.magic import register_line_magic, register_cell_magic

        @register_cell_magic(base_name)
        @process_line_arguments
        def query_magic(statement, variable=None, show='head', template=True, name=None, auto_transpose=True, **kwargs):

            ip = get_ipython()

            # If name is specified, save this query as a template that can be used in subsequent queries
            if name is not None:
                if not hasattr(self, '_templates'):
                    self._templates = {}
                self._templates[name] = statement

            # Create the query by inserting stored templates where indicated by `::<name>::`
            # And then render the template using variables in the current user namespace.
            if template:
                statement = Template(statement,
                                     block_start_string='#{%#',
                                     block_end_string='#%}#',
                                     variable_start_string='::',
                                     variable_end_string='::',
                                     comment_start_string='#{%#',
                                     comment_end_string='#%}#').render(getattr(self, '_templates', {}))
                statement = Template(statement).render(ip.user_ns)

            result = self(statement, **kwargs)

            if variable is not None:
                ip.user_ns[variable] = result

            if kwargs.get('render_only', False):
                print(result, file=sys.stderr)
                return
            elif variable is None:
                return result

            if isinstance(show, int):
                r = result.head(show)
                if show <= 10:
                    r = r.T
                return r
            elif show == 'head':
                r = result.head()
            elif show == 'all':
                r = result
            elif show == 'none':
                return None
            else:
                raise ValueError("Omniduct does not recognise the argument show='{0}' in cell magic.".format(show))

            if auto_transpose and len(r) <= 10:
                return r.T
            return r

        @register_line_magic("{}.{}".format(base_name, 'desc'))
        @process_line_arguments
        def table_desc(table_name, **kwargs):
            return self.table_desc(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'head'))
        @process_line_arguments
        def table_desc(table_name, **kwargs):
            return self.table_head(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'props'))
        @process_line_arguments
        def table_desc(table_name, **kwargs):
            return self.table_props(table_name, **kwargs)



from .caches.local import LocalCache
from .databases.hiveserver2 import HiveServer2Client
from .databases.presto import PrestoClient
from .databases.sqlalchemy import SQLAlchemyClient
from .filesystems.local import LocalFsClient
from .filesystems.webhdfs import WebHdfsClient
from .remotes.ssh import SSHClient

# from .remotes.ssh_paramiko import ParamikoSSHClient  # Not yet ready for prime time



import io
import os
from abc import abstractmethod

from omniduct.duct import Duct
from omniduct.utils.magics import MagicsProvider, process_line_arguments


class FileSystemClient(Duct, MagicsProvider):

    DUCT_TYPE = Duct.Type.FILESYSTEM
    DEFAULT_PORT = None

    def __init__(self, *args, **kwargs):
        '''
        This is a shim __init__ function that passes all arguments onto
        `self._init`, which is implemented by subclasses. This allows subclasses
        to instantiate themselves with arbitrary parameters.
        '''
        Duct.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)
        self._init(*args, **kwargs)

    @abstractmethod
    def _init(self):
        pass

    # Filesystem accessors

    def exists(self, path):
        return self.connect()._exists(path)

    @abstractmethod
    def _exists(self, path):
        raise NotImplementedError

    def isdir(self, path):
        return self.connect()._isdir(path)

    @abstractmethod
    def _isdir(self, path):
        raise NotImplementedError

    def isfile(self, path):
        return self.connect()._isfile(path)

    @abstractmethod
    def _isfile(self, path):
        raise NotImplementedError

    def listdir(self, path=None):
        return self.connect()._listdir(path)

    @abstractmethod
    def _listdir(self, path):
        raise NotImplementedError

    def showdir(self, path=None):
        return self.connect()._showdir(path)

    @abstractmethod
    def _showdir(self, path):
        raise NotImplementedError

    # Directory handling

    def mkdir(self, path, parents=True):
        pass

    # File handling

    def open(self, path, mode='rt'):
        return FileSystemFile(self, path, mode)

    def _file_read(self, path, size=-1, offset=0, binary=False):
        return self.connect()._file_read_(path, size=size, binary=binary)

    @abstractmethod
    def _file_read_(self, path, size=-1, offset=0, binary=False):
        raise NotImplementedError

    def _file_write(self, path, s, binary=False):
        return self.connect()._file_write_(path, s, binary)

    @abstractmethod
    def _file_write_(self, path, s, binary):
        raise NotImplementedError

    def _file_append(self, path, s, binary=False):
        return self.connect()._file_append_(path, s, binary)

    @abstractmethod
    def _file_append_(self, path, s, binary):
        raise NotImplementedError

    # File transfer

    def copy_to_local(self, source, dest=None, overwrite=False):
        '''
        Copies a file from `source` on the remote host to `dest` on the local host.
        If `overwrite` is `True`, overwrites file on the local host.
        '''
        if dest is None:
            dest = os.path.basename(source)
        if not overwrite and os.path.exists(dest):
            raise RuntimeError("File already exists on filesystem.")
        return self.connect()._copy_to_local(source, dest, overwrite)

    def copy_from_local(self, source, dest=None, overwrite=False):
        '''
        Copies a file from `source` on the local host to `dest` on the remote host.
        If `overwrite` is `True`, overwrites file on the remote host.
        '''
        if dest is None:
            dest = os.path.basename(source)
        if not overwrite and self.exists(dest):
            raise RuntimeError("File already exists on filesystem.")
        return self.connect()._copy_from_local(source, dest, overwrite)

    def _copy_to_local(self, source, dest, overwrite):
        """
        A default implementation using other FileSystemClient methods.
        """
        if not self.isfile(source):
            raise ValueError("Only individual file transfers are support at this time.")
        with open(source, 'w') as f:
            return f.write(self._file_read(source, binary=False))

    def _copy_from_local(self, source, dest, overwrite):
        """
        A default implementation using other FileSystemClient methods.
        """
        if not os.path.isfile(source):
            raise ValueError("Only individual file transfers are support at this time.")
        with open(source) as f:
            return self._file_write(dest, f.read())

    # Magics

    def _register_magics(self, base_name):
        from IPython.core.magic import register_line_magic, register_cell_magic

        @register_line_magic("{}.listdir".format(base_name))
        @process_line_arguments
        def listdir(path=''):
            return self.listdir(path)

        @register_line_magic("{}.showdir".format(base_name))
        @process_line_arguments
        def showdir(path=''):
            return self.showdir(path)

        @register_line_magic("{}.read".format(base_name))
        @process_line_arguments
        def read_file(path):
            return self._file_read(path)

        @register_cell_magic("{}.write".format(base_name))
        @process_line_arguments
        def write_file(cell, path):
            return self._file_write(path, cell)


# TODO: properly implement file modes and raise correct errors (consistent with other file types)
class FileSystemFile(object):

    def __init__(self, fs, path, mode='r'):
        self.fs = fs
        self.path = path
        self.mode = mode
        self.offset = 0
        self.closed = False
        self.__modified = False

        if self.binary_mode:
            self.__io_buffer = io.BytesIO()
        else:
            self.__io_buffer = io.StringIO()

        if 'w' not in self.mode:
            self.__io_buffer.write(self.fs._file_read(self.path, binary=self.binary_mode))
            if not self.appending:
                self.__io_buffer.seek(0)

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        try:
            assert len(set(mode)) == len(mode)
            assert sum(l in mode for l in ['r', 'w', 'a', '+', 't', 'b']) == len(mode)
            assert sum(l in mode for l in ['r', 'w', 'a']) == 1
            assert sum(l in mode for l in ['t', 'b']) < 2
        except AssertionError:
            raise ValueError("invalid mode: '{}'".format(mode))
        self.__mode = mode

    @property
    def readable(self):
        return 'r' in self.mode or '+' in self.mode

    @property
    def writeable(self):
        return 'w' in self.mode or 'a' in self.mode or '+' in self.mode

    @property
    def appending(self):
        return 'a' in self.mode

    @property
    def binary_mode(self):
        return 'b' in self.mode

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def close(self):
        self.flush()
        self.closed = True

    def flush(self):
        if not self.writeable or not self.__modified:
            return

        # For the time being, just write out entire buffer. We can consider something cleverer later.
        offset = self.__io_buffer.tell()
        self.__io_buffer.seek(0)
        self.fs._file_write(self.path, self.__io_buffer.read())
        self.__io_buffer.seek(offset)

        self.__modified = False

    def isatty(self):
        return self.__io_buffer.isatty()

    def read(self, size=-1):
        if not self.readable:
            raise io.UnsupportedOperation("File not open for reading")
        return self.__io_buffer.read(size)

    def readline(self, size=-1):
        if not self.readable:
            raise io.UnsupportedOperation("File not open for reading")
        return self.__io_buffer.readline(size)

    def readlines(self, hint=-1):
        if not self.readable:
            raise io.UnsupportedOperation("File not open for reading")
        return self.__io_buffer.readlines(hint)

    def seek(self, pos, whence=0):
        self.__io_buffer.seek(pos)

    def tell(self):
        return self.__io_buffer.tell()

    def write(self, s):
        if not self.writeable:
            raise io.UnsupportedOperation("File not open for writing")
        self.__io_buffer.write(s)
        self.__modified = True



import os
import sys

__all__ = ['__author__', '__author_email__', '__version__']

__author__ = "Matthew Wardrop, Dan Frank"
__author_email__ = "matthew.wardrop@airbnb.com, dan.frank@airbnb.com"
__version__ = "0.2.6"


# These are the core dependencies, and should not include those which are used only in handling specific protocols.
# Order matters since installation happens from the end of the list
__dependencies__ = [
    "future",  # Python 2/3 support
    "six",  # Python 2/3 support
    "enum34",  # Python 3.4+ style enums in older versions of python

    "pyyaml",  # YAML configuration parsing
    "decorator",  # Decorators used by caching routines
    "progressbar2",  # Support for progressbars in logging routines

    # Database querying libraries
    "jinja2",  # Templating support in databases
    "pandas>=0.17.1",  # Various results including database queries are returned as pandas dataframes
    "sqlparse",  # Neatening of SQL based queries (mainly to avoid missing the cache)
]

PY2 = sys.version_info[0] == 2
if os.name == 'posix' and PY2:
    __dependencies__.append('subprocess32')  # Python 3.2+ subprocess handling for Python 2

__optional_dependencies__ = {
    # Databases
    'hiveserver2': [
        'pyhive[Hive]' if PY2 else 'pyhive>=0.3.0[Hive]',  # Primary client
        'impyla>=0.14.0',  # Primary client
    ],

    'presto': [
        'pyhive[Presto]' if PY2 else 'pyhive>=0.3.0[Presto]',  # Primary client
        'sqlalchemy',  # Schema traversal
        'werkzeug',  # Schema traversal
    ],

    'sqlalchemy': [
        'sqlalchemy'  # Primary client
    ],

    # Filesystems
    'webhdfs': [
        'pywebhdfs',  # Primary client
        'requests',  # For rerouting redirect queries to our port-forwarded services
    ],

    # Remotes
    'ssh': [
        'pexpect',  # Command line handling (including smartcard activation)
    ],

    'ssh_paramiko': [
        'paramiko',  # Primary client
        'pexpect',  # Command line handling (including smartcard activation)
    ]
}
__optional_dependencies__['all'] = [dep for deps in __optional_dependencies__.values() for dep in deps]
